# Generated by Django 4.2.24 on 2025-10-01 17:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("lti_tool", "0006_ltiregistration_audience"),
    ]

    operations = [
        migrations.AddField(
            model_name="lticontext",
            name="lti1p1_id_on_platform",
            field=models.CharField(
                blank=True, max_length=255, verbose_name="LTI 1.1 ID"
            ),
        ),
        migrations.AddField(
            model_name="ltiplatforminstance",
            name="lti1p1_id_on_platform",
            field=models.CharField(
                blank=True, max_length=255, verbose_name="LTI 1.1 ID"
            ),
        ),
        migrations.AddField(
            model_name="ltiresourcelink",
            name="lti1p1_id_on_platform",
            field=models.CharField(
                blank=True, max_length=255, verbose_name="LTI 1.1 ID"
            ),
        ),
        migrations.AddField(
            model_name="ltiuser",
            name="lti1p1_id_on_platform",
            field=models.CharField(
                blank=True, max_length=255, verbose_name="LTI 1.1 ID"
            ),
        ),
    ]



import json

import pytest

from lti_tool import models, views


@pytest.mark.django_db
def test_jwks(rf):
    models.Key.objects.generate()
    request = rf.get("/jwks.json")
    response = views.jwks(request)
    assert response.status_code == 200
    jwks = json.loads(response.content)
    assert "keys" in jwks
    assert len(jwks["keys"]) == 1



# Generated by Django 4.0.2 on 2022-02-26 19:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('lti_tool', '0002_create_lti_domain_models'),
    ]

    operations = [
        migrations.AddField(
            model_name='lticontext',
            name='memberships_url',
            field=models.URLField(blank=True),
        ),
        migrations.AddField(
            model_name='ltimembership',
            name='is_active',
            field=models.BooleanField(default=True, verbose_name='is active'),
        ),
    ]



from datetime import timedelta
from io import StringIO

from django.core.management import call_command
from django.utils.timezone import now

import pytest

from lti_tool.models import Key


@pytest.mark.django_db
def test_rotate_keys():
    Key.objects.create(
        public_key="", private_key="", datetime_created=now() - timedelta(days=1)
    )
    Key.objects.create(
        public_key="", private_key="", datetime_created=now() - timedelta(days=3)
    )
    out = StringIO()
    call_command("rotate_keys", "--deactivate-after=2", stdout=out)
    assert "Created new key" in out.getvalue()
    assert "Deactivated 1 key." in out.getvalue()
    assert Key.objects.count() == 3
    assert Key.objects.active().count() == 2



from pylti1p3.exception import LtiException

from .constants import SESSION_KEY
from .models import AbsentLtiLaunch
from .utils import get_launch_from_request


class LtiLaunchMiddleware:
    def __init__(self, get_response) -> None:
        self.get_response = get_response

    def __call__(self, request):
        launch_id = request.session.get(SESSION_KEY)
        try:
            request.lti_launch = get_launch_from_request(request, launch_id)
        except LtiException:
            request.lti_launch = AbsentLtiLaunch()
        return self.get_response(request)



import factory

from . import models


class LtiRegistrationFactory(factory.django.DjangoModelFactory):
    """Factory for LtiRegistration objects."""

    name = factory.Faker("company")
    issuer = factory.Faker("url")
    client_id = factory.Faker("pystr")
    auth_url = factory.Faker("uri")
    token_url = factory.Faker("uri")
    keyset_url = factory.Faker("uri")

    class Meta:
        model = models.LtiRegistration


class LtiPlatformInstanceFactory(factory.django.DjangoModelFactory):
    """Factory for LtiPlatformInstance objects."""

    issuer = factory.Faker("url")
    guid = factory.Faker("pystr")
    contact_email = factory.Faker("email")
    description = factory.Faker("text")
    name = factory.Faker("company")
    url = factory.Faker("uri")
    product_family_code = factory.Faker("word")
    version = factory.Faker("word")

    class Meta:
        model = models.LtiPlatformInstance


class LtiDeploymentFactory(factory.django.DjangoModelFactory):
    """Factory for LtiDeployment objects."""

    registration = factory.SubFactory(LtiRegistrationFactory)
    deployment_id = factory.Faker("pystr")
    platform_instance = factory.SubFactory(
        LtiPlatformInstanceFactory,
        issuer=factory.SelfAttribute("..registration.issuer"),
    )

    class Meta:
        model = models.LtiDeployment


class LtiUserFactory(factory.django.DjangoModelFactory):
    """Factory for LtiUser objects."""

    registration = factory.SubFactory(LtiRegistrationFactory)
    sub = factory.Faker("pystr")
    given_name = factory.Faker("first_name")
    family_name = factory.Faker("last_name")
    name = factory.LazyAttribute(lambda o: f"{o.given_name} {o.family_name}")
    email = factory.Faker("email")
    picture_url = factory.Faker("uri")

    class Meta:
        model = models.LtiUser


class LtiContextFactory(factory.django.DjangoModelFactory):
    """Factory for LtiContext objects."""

    deployment = factory.SubFactory(LtiDeploymentFactory)
    id_on_platform = factory.Faker("pystr")
    label = factory.Faker("word")
    title = factory.Faker("catch_phrase")
    memberships_url = factory.Faker("uri")

    class Meta:
        model = models.LtiContext


class LtiMembershipFactory(factory.django.DjangoModelFactory):
    """Factory for LtiMembership objects."""

    context = factory.SubFactory(LtiContextFactory)
    user = factory.SubFactory(
        LtiUserFactory,
        registration=factory.SelfAttribute("..context.deployment.registration"),
    )

    class Meta:
        model = models.LtiMembership


class LtiResourceLinkFactory(factory.django.DjangoModelFactory):
    """Factory for LtiResourceLink objects."""

    context = factory.SubFactory(LtiContextFactory)
    id_on_platform = factory.Faker("pystr")
    title = factory.Faker("catch_phrase")
    description = factory.Faker("sentence")

    class Meta:
        model = models.LtiResourceLink



import pytest

from lti_tool import factories, models, utils
from lti_tool.lti_core.utils import compute_oauth_consumer_key_sign


@pytest.mark.django_db
class TestSyncUserFromLaunch:
    """Tests for utils.sync_user_from_launch."""

    def test_sync_new_user(self, monkeypatch):
        launch_data = {
            "sub": "abc123",
            "given_name": "First",
            "family_name": "Last",
            "name": "First Last",
            "email": "first.last@example.com",
            "picture": "https://example.com/picture.jpg",
        }
        registration = factories.LtiRegistrationFactory()
        monkeypatch.setattr(
            models.LtiLaunch, "get_launch_data", lambda self: launch_data
        )
        monkeypatch.setattr(models.LtiLaunch, "registration", registration)
        lti_launch = models.LtiLaunch(None)
        user = utils.sync_user_from_launch(lti_launch)
        assert user.sub == "abc123"
        assert user.given_name == "First"
        assert user.family_name == "Last"
        assert user.name == "First Last"
        assert user.email == "first.last@example.com"
        assert user.picture_url == "https://example.com/picture.jpg"

    def test_sync_existing_user(self, monkeypatch):
        registration = factories.LtiRegistrationFactory()
        factories.LtiUserFactory(registration=registration, sub="abc123")
        launch_data = {
            "sub": "abc123",
            "given_name": "First",
            "family_name": "Last",
            "name": "First Last",
            "email": "first.last@example.com",
            "picture": "https://example.com/picture.jpg",
        }
        monkeypatch.setattr(
            models.LtiLaunch, "get_launch_data", lambda self: launch_data
        )
        monkeypatch.setattr(models.LtiLaunch, "registration", registration)
        lti_launch = models.LtiLaunch(None)
        user = utils.sync_user_from_launch(lti_launch)
        assert models.LtiUser.objects.count() == 1
        assert user.sub == "abc123"
        assert user.given_name == "First"
        assert user.family_name == "Last"
        assert user.name == "First Last"
        assert user.email == "first.last@example.com"
        assert user.picture_url == "https://example.com/picture.jpg"

    @pytest.mark.parametrize("user_id_present", [True, False])
    def test_sync_user_with_migration(self, monkeypatch, user_id_present):
        migration_claim = {"oauth_consumer_key": "179248902"}
        if user_id_present:
            migration_claim["user_id"] = "lti1p1"
        migration_claim["oauth_consumer_key_sign"] = compute_oauth_consumer_key_sign(
            "179248902",
            "689302",
            "https://lmsvendor.com",
            "PM48OJSfGDTAzAo",
            1551290856,
            "172we8671fd8z",
            "oauth_secret",
        )
        launch_data = {
            "nonce": "172we8671fd8z",
            "iat": 1551290796,
            "exp": 1551290856,
            "iss": "https://lmsvendor.com",
            "aud": "PM48OJSfGDTAzAo",
            "https://purl.imsglobal.org/spec/lti/claim/deployment_id": "689302",
            "sub": "abc123",
            "https://purl.imsglobal.org/spec/lti/claim/lti1p1": migration_claim,
        }
        registration = factories.LtiRegistrationFactory()
        monkeypatch.setattr(
            models.LtiLaunch, "get_launch_data", lambda self: launch_data
        )
        monkeypatch.setattr(models.LtiLaunch, "registration", registration)
        lti_launch = models.LtiLaunch(None)
        user = utils.sync_user_from_launch(lti_launch, "oauth_secret")
        assert user.sub == "abc123"
        assert user.lti1p1_id_on_platform == "lti1p1" if user_id_present else "abc123"


@pytest.mark.django_db
class TestSyncContextFromLaunch:
    """Tests for utils.sync_context_from_launch."""

    def test_sync_empty_context(self, monkeypatch):
        context_claim = None
        deployment = factories.LtiDeploymentFactory()
        monkeypatch.setattr(models.LtiLaunch, "context_claim", context_claim)
        monkeypatch.setattr(models.LtiLaunch, "deployment", deployment)
        lti_launch = models.LtiLaunch(None)
        context = utils.sync_context_from_launch(lti_launch)
        assert context.id_on_platform == ""

    def test_sync_new_context(self, monkeypatch):
        context_claim = {
            "id": "a-context-id",
            "title": "A Context Title",
            "label": "CTX101",
            "type": ["http://purl.imsglobal.org/vocab/lis/v2/course#CourseOffering"],
        }
        deployment = factories.LtiDeploymentFactory()
        monkeypatch.setattr(models.LtiLaunch, "context_claim", context_claim)
        monkeypatch.setattr(models.LtiLaunch, "deployment", deployment)
        lti_launch = models.LtiLaunch(None)
        context = utils.sync_context_from_launch(lti_launch)
        assert context.id_on_platform == "a-context-id"
        assert context.label == "CTX101"
        assert context.title == "A Context Title"
        assert context.is_course_offering

    def test_sync_new_context_with_ags(self, monkeypatch):
        context_claim = {"id": "a-context-id"}
        ags_claim = {
            "scope": [
                "https://purl.imsglobal.org/spec/lti-ags/scope/lineitem",
                "https://purl.imsglobal.org/spec/lti-ags/scope/result.readonly",
                "https://purl.imsglobal.org/spec/lti-ags/scope/score",
            ],
            "lineitems": "https://www.example.com/2344/lineitems/",
            "lineitem": "https://www.example.com/2344/lineitems/1234/lineitem",
        }
        deployment = factories.LtiDeploymentFactory()
        monkeypatch.setattr(models.LtiLaunch, "context_claim", context_claim)
        monkeypatch.setattr(models.LtiLaunch, "ags_claim", ags_claim)
        monkeypatch.setattr(models.LtiLaunch, "deployment", deployment)
        lti_launch = models.LtiLaunch(None)
        context = utils.sync_context_from_launch(lti_launch)
        assert context.id_on_platform == "a-context-id"
        assert context.lineitems_url == "https://www.example.com/2344/lineitems/"
        assert not context.can_query_lineitems
        assert context.can_manage_lineitems
        assert context.can_publish_scores
        assert context.can_access_results

    def test_sync_existing_context(self, monkeypatch):
        deployment = factories.LtiDeploymentFactory()
        factories.LtiContextFactory(
            deployment=deployment, id_on_platform="ctx-1", is_group=True
        )
        new_context_claim = {
            "id": "ctx-1",
            "title": "New Context Title",
            "label": "CTX101",
            "type": ["http://purl.imsglobal.org/vocab/lis/v2/course#CourseOffering"],
        }
        monkeypatch.setattr(models.LtiLaunch, "context_claim", new_context_claim)
        monkeypatch.setattr(models.LtiLaunch, "deployment", deployment)
        lti_launch = models.LtiLaunch(None)
        updated_context = utils.sync_context_from_launch(lti_launch)
        assert models.LtiContext.objects.count() == 1
        assert updated_context.title == "New Context Title"
        assert updated_context.label == "CTX101"
        assert not updated_context.is_group
        assert updated_context.is_course_offering

    @pytest.mark.parametrize("context_id_present", [True, False])
    def test_sync_context_with_migration(self, monkeypatch, context_id_present):
        context_claim = {"id": "c1p3"}
        deployment = factories.LtiDeploymentFactory()
        migration_claim = {"oauth_consumer_key": "179248902"}
        if context_id_present:
            migration_claim["context_id"] = "c1p1"
        migration_claim["oauth_consumer_key_sign"] = compute_oauth_consumer_key_sign(
            "179248902",
            deployment.deployment_id,
            "https://lmsvendor.com",
            "PM48OJSfGDTAzAo",
            1551290856,
            "172we8671fd8z",
            "oauth_secret",
        )
        launch_data = {
            "nonce": "172we8671fd8z",
            "iat": 1551290796,
            "exp": 1551290856,
            "iss": "https://lmsvendor.com",
            "aud": "PM48OJSfGDTAzAo",
            "https://purl.imsglobal.org/spec/lti/claim/deployment_id": (
                deployment.deployment_id
            ),
            "https://purl.imsglobal.org/spec/lti/claim/context": context_claim,
            "sub": "abc123",
            "https://purl.imsglobal.org/spec/lti/claim/lti1p1": migration_claim,
        }
        monkeypatch.setattr(
            models.LtiLaunch, "get_launch_data", lambda self: launch_data
        )
        monkeypatch.setattr(models.LtiLaunch, "deployment", deployment)
        lti_launch = models.LtiLaunch(None)
        context = utils.sync_context_from_launch(lti_launch, "oauth_secret")
        assert context.id_on_platform == "c1p3"
        assert context.lti1p1_id_on_platform == "c1p1" if context_id_present else "c1p3"


@pytest.mark.django_db
class TestSyncMembershipFromLaunch:
    """Tests for utils.sync_membership_from_launch."""

    def test_sync_new_membership(self, monkeypatch):
        user = factories.LtiUserFactory()
        context = factories.LtiContextFactory(
            deployment__registration=user.registration
        )
        roles_claim = ["http://purl.imsglobal.org/vocab/lis/v2/membership#Learner"]
        monkeypatch.setattr(models.LtiLaunch, "roles_claim", roles_claim)
        lti_launch = models.LtiLaunch(None)
        membership = utils.sync_membership_from_launch(lti_launch, user, context)
        assert membership.user == user
        assert membership.context == context
        assert membership.is_learner
        assert not membership.is_administrator
        assert not membership.is_content_developer
        assert not membership.is_instructor
        assert not membership.is_mentor

    def test_sync_existing_membership(self, monkeypatch):
        membership = factories.LtiMembershipFactory(is_content_developer=True)
        roles_claim = ["http://purl.imsglobal.org/vocab/lis/v2/membership#Instructor"]
        monkeypatch.setattr(models.LtiLaunch, "roles_claim", roles_claim)
        lti_launch = models.LtiLaunch(None)
        updated_membership = utils.sync_membership_from_launch(
            lti_launch, membership.user, membership.context
        )
        assert models.LtiMembership.objects.count() == 1
        assert updated_membership.is_content_developer
        assert updated_membership.is_instructor


@pytest.mark.django_db
class TestSyncResourceLinkFromLaunch:
    """Tests for utils.sync_resource_link_from_launch."""

    @pytest.mark.parametrize(
        ["claim_title", "claim_description", "title", "description"],
        [("Title", "Description", "Title", "Description"), (None, None, "", "")],
    )
    def test_sync_new_resource_link(
        self, monkeypatch, claim_title, claim_description, title, description
    ):
        context = factories.LtiContextFactory()
        resource_link_claim = {
            "id": "resource-link-id",
            "title": claim_title,
            "description": claim_description,
        }
        monkeypatch.setattr(
            models.LtiLaunch, "resource_link_claim", resource_link_claim
        )
        lti_launch = models.LtiLaunch(None)
        resource_link = utils.sync_resource_link_from_launch(lti_launch, context)
        assert resource_link.id_on_platform == "resource-link-id"
        assert resource_link.title == title
        assert resource_link.description == description

    @pytest.mark.parametrize(
        ["claim_title", "claim_description", "title", "description"],
        [("Title", "Description", "Title", "Description"), (None, None, "", "")],
    )
    def test_sync_existing_resource_link(
        self, monkeypatch, claim_title, claim_description, title, description
    ):
        context = factories.LtiContextFactory()
        resource_link = factories.LtiResourceLinkFactory(context=context)
        resource_link_claim = {
            "id": resource_link.id_on_platform,
            "title": claim_title,
            "description": claim_description,
        }
        monkeypatch.setattr(
            models.LtiLaunch, "resource_link_claim", resource_link_claim
        )
        lti_launch = models.LtiLaunch(None)
        resource_link = utils.sync_resource_link_from_launch(lti_launch, context)
        assert models.LtiResourceLink.objects.count() == 1
        assert resource_link.title == title
        assert resource_link.description == description

    @pytest.mark.parametrize("resource_link_id_present", (True, False))
    def test_sync_resource_link_with_migration(
        self, monkeypatch, resource_link_id_present
    ):
        context = factories.LtiContextFactory()
        resource_link_claim = {
            "id": "rl1p3",
        }
        migration_claim = {"oauth_consumer_key": "179248902"}
        if resource_link_id_present:
            migration_claim["resource_link_id"] = "rl1p1"
        migration_claim["oauth_consumer_key_sign"] = compute_oauth_consumer_key_sign(
            "179248902",
            "deployment",
            "https://lmsvendor.com",
            "PM48OJSfGDTAzAo",
            1551290856,
            "172we8671fd8z",
            "oauth_secret",
        )
        launch_data = {
            "nonce": "172we8671fd8z",
            "iat": 1551290796,
            "exp": 1551290856,
            "iss": "https://lmsvendor.com",
            "aud": "PM48OJSfGDTAzAo",
            "https://purl.imsglobal.org/spec/lti/claim/deployment_id": "deployment",
            "https://purl.imsglobal.org/spec/lti/claim/resource_link": (
                resource_link_claim
            ),
            "sub": "abc123",
            "https://purl.imsglobal.org/spec/lti/claim/lti1p1": migration_claim,
        }
        monkeypatch.setattr(
            models.LtiLaunch, "get_launch_data", lambda self: launch_data
        )
        lti_launch = models.LtiLaunch(None)
        resource_link = utils.sync_resource_link_from_launch(
            lti_launch, context, "oauth_secret"
        )
        assert resource_link.id_on_platform == "rl1p3"
        assert (
            resource_link.lti1p1_id_on_platform == "rl1p1"
            if resource_link_id_present
            else "rl1p3"
        )


@pytest.mark.django_db
class TestSyncPlatformInstanceFromLaunch:
    """Tests for utils.sync_platform_instance_from_launch."""

    def test_no_platform_claim(self, monkeypatch):
        deployment = factories.LtiDeploymentFactory()
        issuer = deployment.registration.issuer
        launch_data = {
            "iss": issuer,
        }
        monkeypatch.setattr(models.LtiLaunch, "get_launch_data", lambda s: launch_data)
        monkeypatch.setattr(models.LtiLaunch, "deployment", deployment)
        lti_launch = models.LtiLaunch(None)
        platform_instance = utils.sync_platform_instance_from_launch(lti_launch)
        assert platform_instance is None

    def test_sync_new_platform_instance(self, monkeypatch):
        deployment = factories.LtiDeploymentFactory(platform_instance=None)
        issuer = deployment.registration.issuer
        launch_data = {
            "iss": issuer,
            "https://purl.imsglobal.org/spec/lti/claim/tool_platform": {
                "guid": "guid",
                "contact_email": "email@example.com",
                "description": "Description",
                "name": "Name",
                "url": "https://www.example.com",
                "product_family_code": "example",
                "version": "1.0",
            },
        }
        monkeypatch.setattr(models.LtiLaunch, "get_launch_data", lambda s: launch_data)
        monkeypatch.setattr(models.LtiLaunch, "deployment", deployment)
        lti_launch = models.LtiLaunch(None)
        platform_instance = utils.sync_platform_instance_from_launch(lti_launch)
        assert platform_instance.issuer == issuer
        assert platform_instance.guid == "guid"
        assert platform_instance.contact_email == "email@example.com"
        assert platform_instance.description == "Description"
        assert platform_instance.name == "Name"
        assert platform_instance.url == "https://www.example.com"
        assert platform_instance.product_family_code == "example"
        assert platform_instance.version == "1.0"
        assert platform_instance.deployments.first() == deployment

    def test_sync_existing_platform_instance(self, monkeypatch):
        deployment = factories.LtiDeploymentFactory()
        platform_instance = deployment.platform_instance
        launch_data = {
            "iss": platform_instance.issuer,
            "https://purl.imsglobal.org/spec/lti/claim/tool_platform": {
                "guid": platform_instance.guid,
                "contact_email": "email@example.com",
                "description": "Description",
                "name": "Name",
                "url": "https://www.example.com",
                "product_family_code": "example",
                "version": "1.0",
            },
        }
        monkeypatch.setattr(models.LtiLaunch, "get_launch_data", lambda s: launch_data)
        monkeypatch.setattr(models.LtiLaunch, "deployment", deployment)
        lti_launch = models.LtiLaunch(None)
        updated_platform_instance = utils.sync_platform_instance_from_launch(lti_launch)
        assert models.LtiPlatformInstance.objects.count() == 1
        assert updated_platform_instance.guid == platform_instance.guid
        assert updated_platform_instance.contact_email == "email@example.com"
        assert updated_platform_instance.description == "Description"
        assert updated_platform_instance.name == "Name"
        assert updated_platform_instance.url == "https://www.example.com"
        assert updated_platform_instance.product_family_code == "example"
        assert updated_platform_instance.version == "1.0"
        assert updated_platform_instance.deployments.first() == deployment

    @pytest.mark.parametrize("platform_instance_id_present", (True, False))
    def test_sync_platform_instance_with_migration(
        self, monkeypatch, platform_instance_id_present
    ):
        deployment = factories.LtiDeploymentFactory(platform_instance=None)
        issuer = deployment.registration.issuer
        migration_claim = {"oauth_consumer_key": "179248902"}
        if platform_instance_id_present:
            migration_claim["tool_consumer_instance_guid"] = "pi1p1"
        migration_claim["oauth_consumer_key_sign"] = compute_oauth_consumer_key_sign(
            "179248902",
            deployment.deployment_id,
            issuer,
            "PM48OJSfGDTAzAo",
            1551290856,
            "172we8671fd8z",
            "oauth_secret",
        )
        launch_data = {
            "iss": issuer,
            "https://purl.imsglobal.org/spec/lti/claim/tool_platform": {
                "guid": "pi1p3"
            },
            "nonce": "172we8671fd8z",
            "iat": 1551290796,
            "exp": 1551290856,
            "aud": "PM48OJSfGDTAzAo",
            "https://purl.imsglobal.org/spec/lti/claim/deployment_id": (
                deployment.deployment_id
            ),
            "sub": "abc123",
            "https://purl.imsglobal.org/spec/lti/claim/lti1p1": migration_claim,
        }
        monkeypatch.setattr(models.LtiLaunch, "get_launch_data", lambda s: launch_data)
        monkeypatch.setattr(models.LtiLaunch, "deployment", deployment)
        lti_launch = models.LtiLaunch(None)
        platform_instance = utils.sync_platform_instance_from_launch(
            lti_launch, "oauth_secret"
        )
        assert platform_instance.guid == "pi1p3"
        assert (
            platform_instance.lti1p1_id_on_platform == "pi1p1"
            if platform_instance_id_present
            else "pi1p3"
        )


@pytest.mark.django_db
class TestDjangoToolConfig:
    """Tests for utils.DjangoToolConfig."""

    @pytest.mark.parametrize("is_known_deployment", (True, False))
    def test_find_deployment(self, is_known_deployment):
        models.Key.objects.generate()
        registration = factories.LtiRegistrationFactory()
        deployment_id = "a-deployment"
        if is_known_deployment:
            factories.LtiDeploymentFactory(
                registration=registration, deployment_id=deployment_id, is_active=True
            )

        tool_config = utils.DjangoToolConfig(registration_uuid=registration.uuid)
        tool_config.find_registration_by_issuer(registration.issuer)
        deployment = tool_config.find_deployment(registration.issuer, deployment_id)

        assert deployment.get_deployment_id() == deployment_id
        assert tool_config.deployment.deployment_id == deployment_id
        assert tool_config.deployment.is_active == is_known_deployment
        assert models.LtiDeployment.objects.count() == 1

    @pytest.mark.parametrize("is_known_deployment", (True, False))
    def test_find_deployment_by_params(self, is_known_deployment):
        models.Key.objects.generate()
        registration = factories.LtiRegistrationFactory()
        deployment_id = "a-deployment"
        if is_known_deployment:
            factories.LtiDeploymentFactory(
                registration=registration, deployment_id=deployment_id, is_active=True
            )

        tool_config = utils.DjangoToolConfig()
        tool_config.find_registration_by_params(
            registration.issuer, registration.client_id
        )
        deployment = tool_config.find_deployment_by_params(
            registration.issuer, deployment_id, registration.client_id
        )

        assert deployment.get_deployment_id() == deployment_id
        assert tool_config.deployment.deployment_id == deployment_id
        assert tool_config.deployment.is_active == is_known_deployment
        assert models.LtiDeployment.objects.count() == 1



from typing import Optional

from django.http import (
    HttpRequest,
    HttpResponse,
    HttpResponseBadRequest,
    HttpResponseForbidden,
    JsonResponse,
)
from django.http.response import HttpResponseRedirect
from django.utils.decorators import method_decorator
from django.utils.translation import gettext as _
from django.views import View
from django.views.decorators.clickjacking import xframe_options_exempt
from django.views.decorators.csrf import csrf_exempt

from pylti1p3.contrib.django import DjangoCacheDataStorage, DjangoOIDCLogin

from lti_tool.utils import DjangoToolConfig, get_launch_from_request

from .constants import SESSION_KEY
from .models import Key, LtiLaunch
from .types import LtiHttpRequest
from .utils import sync_data_from_launch


def jwks(request):
    """Makes a JWKS available to LTI platforms."""
    return JsonResponse(Key.objects.as_jwks())


@method_decorator((csrf_exempt, xframe_options_exempt), name="dispatch")
class OIDCLoginInitView(View):
    """Handles OIDC 3rd-party login initiation for an LTI launch."""

    main_msg = (
        "Your browser prevents embedded content from using cookies.  To work "
        "around this, the content must be opened in a new tab or window.  "
    )
    click_msg = "Open a new tab or window now."
    loading_msg = "Loading..."

    def get(self, request, *args, **kwargs):
        registration_uuid = kwargs.get("registration_uuid")
        return self.get_oidc_response(request, registration_uuid, request.GET)

    def post(self, request, *args, **kwargs):
        registration_uuid = kwargs.get("registration_uuid")
        return self.get_oidc_response(request, registration_uuid, request.POST)

    def get_redirect_url(self, target_link_uri: str) -> str:
        """Returns the redirect_uri to use for the OIDC initiation response."""
        return target_link_uri

    def get_oidc_response(self, request, registration_uuid, params):
        tool_conf = DjangoToolConfig(registration_uuid)
        launch_data_storage = DjangoCacheDataStorage()
        oidc_login = DjangoOIDCLogin(
            request, tool_conf, launch_data_storage=launch_data_storage
        )
        target_link_uri = params.get("target_link_uri")
        if target_link_uri is None:
            return HttpResponseBadRequest("Missing target_link_uri parameter.")
        redirect_url = self.get_redirect_url(target_link_uri)
        return oidc_login.enable_check_cookies(
            self.main_msg, self.click_msg, self.loading_msg
        ).redirect(redirect_url)


@method_decorator(csrf_exempt, name="dispatch")
class LtiLaunchBaseView(View):
    """Base view for handling LTI 1.3 launches.

    Inherit from this view and implement the `handle_resource_launch` method. Other
    message types can be handled as well by overriding the appropriate method.
    """

    def post(self, request: LtiHttpRequest, *args, **kwargs):
        request.session.clear()
        lti_launch = get_launch_from_request(request)
        lti1p1_secret = self.get_lti1p1_secret(lti_launch.lti1p1_consumer_key)
        sync_data_from_launch(lti_launch, lti1p1_secret)
        self.launch_setup(request, lti_launch)
        if not lti_launch.deployment.is_active:
            return self.handle_inactive_deployment(request, lti_launch)
        request.session[SESSION_KEY] = lti_launch.get_launch_id()
        request.lti_launch = lti_launch
        if request.lti_launch.is_resource_launch:
            return self.handle_resource_launch(request, lti_launch)
        if request.lti_launch.is_deep_link_launch:
            return self.handle_deep_linking_launch(request, lti_launch)
        if request.lti_launch.is_submission_review_launch:
            return self.handle_submission_review_launch(request, lti_launch)
        if request.lti_launch.is_data_privacy_launch:
            return self.handle_data_privacy_launch(request, lti_launch)

    def get_lti1p1_secret(self, oauth_consumer_key: Optional[str]):
        """Returns the LTI 1.1 secret for the provided consumer key.

        Override this method to validate and sync data from the LTI 1.1 migration claim.
        """
        return None

    def launch_setup(self, request: HttpRequest, lti_launch: LtiLaunch) -> None:
        pass

    def handle_inactive_deployment(
        self, request: HttpRequest, lti_launch: LtiLaunch
    ) -> HttpResponse:
        """Handles LTI launches with inactive deployments."""
        error_msg = _("This deployment is not active.")
        return_url = lti_launch.get_return_url(lti_errormsg=error_msg)
        if return_url is None:
            return HttpResponseForbidden(error_msg)
        return HttpResponseRedirect(return_url)

    def handle_resource_launch(
        self, request: HttpRequest, lti_launch: LtiLaunch
    ) -> HttpResponse:
        """Handles an LTI resource launch.

        Implement this method to handle LTI resource link launch requests,
        as described in the LTI 1.3 core spec at
        https://www.imsglobal.org/spec/lti/v1p3/#resource-link-launch-request-message
        """
        raise NotImplementedError(
            "Subclasses of LtiLaunchBaseView must implement handle_resource_launch"
        )

    def handle_deep_linking_launch(
        self, request: HttpRequest, lti_launch: LtiLaunch
    ) -> HttpResponse:
        error_msg = _("Deep linking is not supported.")
        # TODO: Update once pylti1p3 supports error_msg claim.
        return HttpResponseForbidden(error_msg)

    def handle_submission_review_launch(
        self, request: HttpRequest, lti_launch: LtiLaunch
    ) -> HttpResponse:
        error_msg = _("Submission review launch is not supported.")
        return_url = lti_launch.get_return_url(lti_errormsg=error_msg)
        if return_url is None:
            return HttpResponseForbidden(error_msg)
        return HttpResponseRedirect(return_url)

    def handle_data_privacy_launch(
        self, request: HttpRequest, lti_launch: LtiLaunch
    ) -> HttpResponse:
        error_msg = _("Data privacy launch is not supported.")
        return_url = lti_launch.get_return_url(lti_errormsg=error_msg)
        if return_url is None:
            return HttpResponseForbidden(error_msg)
        return HttpResponseRedirect(return_url)



from enum import Enum

CONTEXT_ROLE_PATTERN = "http://purl.imsglobal.org/vocab/lis/v2/membership#{}"


class ContextRole(str, Enum):
    ADMINISTRATOR = CONTEXT_ROLE_PATTERN.format("Administrator")
    CONTENT_DEVELOPER = CONTEXT_ROLE_PATTERN.format("ContentDeveloper")
    INSTRUCTOR = CONTEXT_ROLE_PATTERN.format("Instructor")
    LEARNER = CONTEXT_ROLE_PATTERN.format("Learner")
    MENTOR = CONTEXT_ROLE_PATTERN.format("Mentor")

    @property
    def short_name(self) -> str:
        """Return the short name of this role."""
        return self.value[len(CONTEXT_ROLE_PATTERN.format("")) :]

    @property
    def full_name(self) -> str:
        """Return the full name of this role."""
        return self.value


CONTEXT_TYPE_PATTERN = "http://purl.imsglobal.org/vocab/lis/v2/course#{}"


class ContextType(str, Enum):
    COURSE_TEMPLATE = CONTEXT_TYPE_PATTERN.format("CourseTemplate")
    COURSE_OFFERING = CONTEXT_TYPE_PATTERN.format("CourseOffering")
    COURSE_SECTION = CONTEXT_TYPE_PATTERN.format("CourseSection")
    GROUP = CONTEXT_TYPE_PATTERN.format("Group")

    @property
    def short_name(self) -> str:
        """Return the short name of this context."""
        return self.value[len(CONTEXT_TYPE_PATTERN.format("")) :]

    @property
    def full_name(self) -> str:
        """Return the full name of this context."""
        return self.value


SYSTEM_ROLE_PATTERN = "http://purl.imsglobal.org/vocab/lis/v2/system/person#{}"


class SystemRole(str, Enum):
    ADMINISTRATOR = SYSTEM_ROLE_PATTERN.format("Administrator")
    NONE = SYSTEM_ROLE_PATTERN.format("None")
    ACCOUNT_ADMIN = SYSTEM_ROLE_PATTERN.format("AccountAdmin")
    CREATOR = SYSTEM_ROLE_PATTERN.format("Creator")
    SYS_ADMIN = SYSTEM_ROLE_PATTERN.format("SysAdmin")
    SYS_SUPPORT = SYSTEM_ROLE_PATTERN.format("SysSupport")
    USER = SYSTEM_ROLE_PATTERN.format("User")

    @property
    def short_name(self) -> str:
        """Return the short name of this role."""
        return self.value[len(SYSTEM_ROLE_PATTERN.format("")) :]

    @property
    def full_name(self) -> str:
        """Return the full name of this roles."""
        return self.value


INSTITUTION_ROLE_PATTERN = (
    "http://purl.imsglobal.org/vocab/lis/v2/institution/person#{}"
)


class InstitutionRole(str, Enum):
    ADMINISTRATOR = INSTITUTION_ROLE_PATTERN.format("Administrator")
    FACULTY = INSTITUTION_ROLE_PATTERN.format("Faculty")
    GUEST = INSTITUTION_ROLE_PATTERN.format("Guest")
    NONE = INSTITUTION_ROLE_PATTERN.format("None")
    OTHER = INSTITUTION_ROLE_PATTERN.format("Other")
    STAFF = INSTITUTION_ROLE_PATTERN.format("Staff")
    STUDENT = INSTITUTION_ROLE_PATTERN.format("Student")
    ALUMNI = INSTITUTION_ROLE_PATTERN.format("Alumni")
    INSTRUCTOR = INSTITUTION_ROLE_PATTERN.format("Instructor")
    LEARNER = INSTITUTION_ROLE_PATTERN.format("Learner")
    MEMBER = INSTITUTION_ROLE_PATTERN.format("Member")
    MENTOR = INSTITUTION_ROLE_PATTERN.format("Mentor")
    OBSERVER = INSTITUTION_ROLE_PATTERN.format("Observer")
    PROSPECTIVE_STUDENT = INSTITUTION_ROLE_PATTERN.format("ProspectiveStudent")

    @property
    def short_name(self) -> str:
        """Return the short name of this role."""
        return self.value[len(INSTITUTION_ROLE_PATTERN.format("")) :]

    @property
    def full_name(self) -> str:
        """Return the full name of this roles."""
        return self.value


class AgsScope(str, Enum):
    MANAGE_LINEITEMS = "https://purl.imsglobal.org/spec/lti-ags/scope/lineitem"
    QUERY_LINEITEMS = "https://purl.imsglobal.org/spec/lti-ags/scope/lineitem.readonly"
    PUBLISH_SCORES = "https://purl.imsglobal.org/spec/lti-ags/scope/score"
    ACCESS_RESULTS = "https://purl.imsglobal.org/spec/lti-ags/scope/result.readonly"



from django.contrib import admin
from django.utils.translation import gettext as _

from . import models


class LtiDeploymentInline(admin.TabularInline):
    model = models.LtiDeployment


@admin.register(models.LtiRegistration)
class LtiRegistrationAdmin(admin.ModelAdmin):
    list_display = ["name", "issuer", "client_id", "datetime_created", "is_active"]
    fieldsets = [
        (
            None,
            {
                "fields": [
                    "name",
                    "uuid",
                    "issuer",
                    "client_id",
                    "audience",
                    "auth_url",
                    "token_url",
                    "keyset_url",
                    "is_active",
                ]
            },
        ),
        (
            _("Static keys"),
            {
                "description": _("For platforms not supporting JWK Set use by tools."),
                "fields": ["public_key", "private_key"],
                "classes": ["collapse"],
            },
        ),
    ]
    inlines = [LtiDeploymentInline]



# Generated by Django 4.0.8 on 2023-05-18 19:27

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("lti_tool", "0004_lticontext_can_access_results_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="LtiLineItem",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("url", models.URLField(unique=True, verbose_name="URL on platform")),
                ("maximum_score", models.FloatField(verbose_name="maximum score")),
                ("label", models.CharField(max_length=255, verbose_name="label")),
                (
                    "tag",
                    models.CharField(blank=True, max_length=255, verbose_name="tag"),
                ),
                (
                    "resource_id",
                    models.CharField(
                        blank=True, max_length=255, verbose_name="resource ID"
                    ),
                ),
                (
                    "start_datetime",
                    models.DateTimeField(null=True, verbose_name="start date/time"),
                ),
                (
                    "end_datetime",
                    models.DateTimeField(null=True, verbose_name="end date/time"),
                ),
                (
                    "context",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="line_items",
                        to="lti_tool.lticontext",
                        verbose_name="context",
                    ),
                ),
                (
                    "resource_link",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="lti_tool.ltiresourcelink",
                    ),
                ),
            ],
        ),
    ]



import functools

from django.core.exceptions import PermissionDenied
from django.utils.translation import gettext as _


def lti_launch_required(view_func):
    """Decorator that requires a session associated with a LTI launch."""

    @functools.wraps(view_func)
    def wrapper(request, *args, **kwargs):
        if request.lti_launch.is_absent:
            raise PermissionDenied(
                _("This page may only be accessed through a LTI launch.")
            )
        return view_func(request, *args, **kwargs)

    return wrapper



# Generated by Django 3.2.9 on 2022-02-08 16:28

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('lti_tool', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='LtiContext',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('id_on_platform', models.CharField(blank=True, max_length=255, verbose_name='ID on platform')),
                ('label', models.CharField(blank=True, max_length=255, verbose_name='label')),
                ('title', models.CharField(blank=True, max_length=500, verbose_name='title')),
                ('is_course_template', models.BooleanField(default=False, verbose_name='is course template')),
                ('is_course_offering', models.BooleanField(default=False, verbose_name='is course offering')),
                ('is_course_section', models.BooleanField(default=False, verbose_name='is course section')),
                ('is_group', models.BooleanField(default=False, verbose_name='is group')),
                ('datetime_created', models.DateTimeField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('datetime_modified', models.DateTimeField(auto_now=True, verbose_name='modified')),
            ],
            options={
                'verbose_name': 'LTI context',
                'verbose_name_plural': 'LTI contexts',
            },
        ),
        migrations.CreateModel(
            name='LtiMembership',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_administrator', models.BooleanField(default=False, verbose_name='is administrator')),
                ('is_content_developer', models.BooleanField(default=False, verbose_name='is_content_developer')),
                ('is_instructor', models.BooleanField(default=False, verbose_name='is instructor')),
                ('is_learner', models.BooleanField(default=False, verbose_name='is learner')),
                ('is_mentor', models.BooleanField(default=False, verbose_name='is mentor')),
                ('datetime_created', models.DateTimeField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('datetime_modified', models.DateTimeField(auto_now=True, verbose_name='modified')),
            ],
            options={
                'verbose_name': 'LTI membership',
                'verbose_name_plural': 'LTI memberships',
            },
        ),
        migrations.CreateModel(
            name='LtiPlatformInstance',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('issuer', models.CharField(max_length=255, verbose_name='issuer')),
                ('guid', models.CharField(max_length=255, verbose_name='GUID')),
                ('contact_email', models.EmailField(blank=True, max_length=254, verbose_name='contact email')),
                ('description', models.TextField(blank=True, verbose_name='description')),
                ('name', models.CharField(blank=True, max_length=500, verbose_name='name')),
                ('url', models.URLField(blank=True, verbose_name='URL')),
                ('product_family_code', models.CharField(blank=True, max_length=500, verbose_name='product family code')),
                ('version', models.CharField(blank=True, max_length=500, verbose_name='version')),
            ],
            options={
                'verbose_name': 'LTI platform instance',
                'verbose_name_plural': 'LTI platform instances',
            },
        ),
        migrations.CreateModel(
            name='LtiUser',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sub', models.CharField(max_length=255, verbose_name='subject')),
                ('given_name', models.CharField(blank=True, max_length=500, verbose_name='given name')),
                ('family_name', models.CharField(blank=True, max_length=500, verbose_name='family name')),
                ('name', models.CharField(blank=True, max_length=500, verbose_name='name')),
                ('email', models.EmailField(blank=True, max_length=254, verbose_name='email')),
                ('picture_url', models.URLField(blank=True, verbose_name='picture URL')),
                ('datetime_created', models.DateTimeField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('datetime_modified', models.DateTimeField(auto_now=True, verbose_name='modified')),
                ('auth_user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='lti_users', to=settings.AUTH_USER_MODEL, verbose_name='auth user')),
                ('registration', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='users', to='lti_tool.ltiregistration', verbose_name='registration')),
            ],
            options={
                'verbose_name': 'LTI user',
                'verbose_name_plural': 'LTI users',
            },
        ),
        migrations.CreateModel(
            name='LtiResourceLink',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('id_on_platform', models.CharField(max_length=255, verbose_name='ID on platform')),
                ('title', models.CharField(blank=True, max_length=500, verbose_name='title')),
                ('description', models.TextField(blank=True, verbose_name='description')),
                ('datetime_created', models.DateTimeField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('datetime_modified', models.DateTimeField(auto_now=True, verbose_name='modified')),
                ('context', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='resource_links', to='lti_tool.lticontext', verbose_name='context')),
            ],
            options={
                'verbose_name': 'LTI resource link',
                'verbose_name_plural': 'LTI resource links',
            },
        ),
        migrations.AddConstraint(
            model_name='ltiplatforminstance',
            constraint=models.UniqueConstraint(fields=('issuer', 'guid'), name='unique_guid_per_issuer'),
        ),
        migrations.AddField(
            model_name='ltimembership',
            name='context',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='memberships', to='lti_tool.lticontext', verbose_name='context'),
        ),
        migrations.AddField(
            model_name='ltimembership',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='memberships', to='lti_tool.ltiuser', verbose_name='user'),
        ),
        migrations.AddField(
            model_name='lticontext',
            name='deployment',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='contexts', to='lti_tool.ltideployment', verbose_name='deployment'),
        ),
        migrations.AddField(
            model_name='lticontext',
            name='members',
            field=models.ManyToManyField(related_name='contexts', through='lti_tool.LtiMembership', to='lti_tool.LtiUser', verbose_name='members'),
        ),
        migrations.AddField(
            model_name='ltideployment',
            name='platform_instance',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='deployments', to='lti_tool.ltiplatforminstance', verbose_name='platform instance'),
        ),
        migrations.AddConstraint(
            model_name='ltiuser',
            constraint=models.UniqueConstraint(fields=('registration', 'sub'), name='unique_sub_per_registration'),
        ),
        migrations.AddConstraint(
            model_name='ltiresourcelink',
            constraint=models.UniqueConstraint(fields=('context', 'id_on_platform'), name='unique_resource_link_id_per_context'),
        ),
        migrations.AddConstraint(
            model_name='ltimembership',
            constraint=models.UniqueConstraint(fields=('user', 'context'), name='no_duplicate_memberships'),
        ),
        migrations.AddConstraint(
            model_name='lticontext',
            constraint=models.UniqueConstraint(fields=('deployment', 'id_on_platform'), name='unique_context_id_per_deployment'),
        ),
    ]



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class LtiToolAppConfig(AppConfig):
    """LTI tool app configuration."""

    name = "lti_tool"
    verbose_name = _("LTI tool")
    default_auto_field = "django.db.models.AutoField"



from datetime import timedelta
from typing import Any

import django.utils.timezone as tz
from django.core.management.base import BaseCommand, CommandParser
from django.db import DatabaseError, transaction
from django.template.defaultfilters import pluralize

from lti_tool.models import Key


class Command(BaseCommand):
    help = "Creates a new keypair and deactivates old ones."

    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument(
            "--deactivate-after",
            type=int,
            default=7,
            help="The age in days beyond which keys should be deactivated.",
        )

    def handle(self, *args: Any, **options: Any) -> None:
        now = tz.now()
        days = timedelta(days=options["deactivate_after"])
        try:
            with transaction.atomic():
                new_key = Key.objects.generate()
                deactivated = (
                    Key.objects.active()
                    .filter(datetime_created__lt=now - days)
                    .update(is_active=False)
                )
            self.stdout.write(self.style.SUCCESS(f"Created new key {new_key}."))
            if deactivated > 0:
                self.stdout.write(
                    self.style.SUCCESS(
                        f"Deactivated {deactivated} key{pluralize(deactivated)}."
                    )
                )
        except DatabaseError:
            self.stdout.write(self.style.ERROR("Unable to rotate keys."))



import hmac
import re
from base64 import b64encode


def normalize_role(role: str) -> str:
    """Expands a simple context role to a full URI, if needed."""
    if re.match(r"^\w+$", role):
        return f"http://purl.imsglobal.org/vocab/lis/v2/membership#{role}"
    return role


def compute_oauth_consumer_key_sign(
    oauth_consumer_key: str,
    deployment_id: str,
    iss: str,
    aud: str,
    exp: int,
    nonce: str,
    oauth_secret: str,
) -> str:
    base_string = "&".join(
        [oauth_consumer_key, deployment_id, iss, aud, str(exp), nonce]
    )
    signed_base_string = hmac.digest(
        oauth_secret.encode(), base_string.encode(), "sha256"
    )
    return b64encode(signed_base_string).decode()


def validate_migration_claim(launch_data: dict, oauth_secret: str) -> bool:
    """Verifies a signature passed in the LTI 1.1 migration claim."""
    try:
        migration_claim = launch_data[
            "https://purl.imsglobal.org/spec/lti/claim/lti1p1"
        ]
        oauth_consumer_key = migration_claim["oauth_consumer_key"]
        oauth_consumer_key_sign = migration_claim["oauth_consumer_key_sign"]
    except KeyError:
        return False
    computed_signature = compute_oauth_consumer_key_sign(
        oauth_consumer_key,
        launch_data["https://purl.imsglobal.org/spec/lti/claim/deployment_id"],
        launch_data["iss"],
        launch_data["aud"],
        launch_data["exp"],
        launch_data["nonce"],
        oauth_secret,
    )
    return hmac.compare_digest(oauth_consumer_key_sign, computed_signature)



# Generated by Django 3.2.9 on 2021-12-13 21:48

from typing import List, Tuple
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies: List[Tuple[str, str]] = []

    operations = [
        migrations.CreateModel(
            name='Key',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('public_key', models.TextField(verbose_name='public key')),
                ('private_key', models.TextField(verbose_name='private key')),
                ('is_active', models.BooleanField(default=True, verbose_name='is active')),
                ('datetime_created', models.DateTimeField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('datetime_modified', models.DateTimeField(auto_now=True, verbose_name='modified')),
            ],
            options={
                'verbose_name': 'JWK',
                'verbose_name_plural': 'JWKs',
                'get_latest_by': 'datetime_created',
            },
        ),
        migrations.CreateModel(
            name='LtiDeployment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('deployment_id', models.CharField(max_length=255, verbose_name='deployment ID')),
                ('is_active', models.BooleanField(default=False, verbose_name='is active')),
                ('datetime_created', models.DateTimeField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('datetime_modified', models.DateTimeField(auto_now=True, verbose_name='modified')),
            ],
            options={
                'verbose_name': 'LTI deployment',
                'verbose_name_plural': 'LTI deployments',
            },
        ),
        migrations.CreateModel(
            name='LtiRegistration',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, verbose_name='name')),
                ('uuid', models.UUIDField(default=uuid.uuid4, verbose_name='UUID')),
                ('issuer', models.CharField(max_length=255, verbose_name='issuer')),
                ('client_id', models.CharField(max_length=255, verbose_name='client ID')),
                ('auth_url', models.URLField(verbose_name='auth URL')),
                ('token_url', models.URLField(verbose_name='access token URL')),
                ('keyset_url', models.URLField(verbose_name='keyset URL')),
                ('is_active', models.BooleanField(default=True, verbose_name='is active')),
                ('public_key', models.TextField(blank=True, verbose_name='public key')),
                ('private_key', models.TextField(blank=True, verbose_name='private key')),
                ('datetime_created', models.DateTimeField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('datetime_modified', models.DateTimeField(auto_now=True, verbose_name='modified')),
            ],
            options={
                'verbose_name': 'LTI registration',
                'verbose_name_plural': 'LTI registrations',
            },
        ),
        migrations.AddConstraint(
            model_name='ltiregistration',
            constraint=models.UniqueConstraint(fields=('issuer', 'client_id'), name='unique_client_id_per_issuer'),
        ),
        migrations.AddField(
            model_name='ltideployment',
            name='registration',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='deployments', to='lti_tool.ltiregistration', verbose_name='registration'),
        ),
        migrations.AddConstraint(
            model_name='ltideployment',
            constraint=models.UniqueConstraint(fields=('registration', 'deployment_id'), name='unique_deployment_id_per_registration'),
        ),
    ]



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("lti_tool", "0005_ltilineitem"),
    ]

    operations = [
        migrations.AddField(
            model_name="ltiregistration",
            name="audience",
            field=models.CharField(blank=True, max_length=255, verbose_name="audience"),
        ),
    ]



from typing import List, cast

from pylti1p3.names_roles import NamesRolesProvisioningService
from pylti1p3.service_connector import ServiceConnector

from lti_tool.models import LtiContext


def fetch_member_data(context: LtiContext) -> List[dict]:
    """Fetches NRPS member data for the given context."""
    if not context.memberships_url:
        return []
    nrps = NamesRolesProvisioningService(
        ServiceConnector(context.deployment.registration.to_registration()),
        {"context_memberships_url": context.memberships_url},
    )
    # TODO: Revisit type casting after pylti1p3 drops support for Python 2.
    return cast(List[dict], nrps.get_members())


def sync_memberships(context: LtiContext):
    """Fetches membership data using NRPS and updates context memberships."""
    member_data = fetch_member_data(context)
    context.update_memberships(member_data)



import json
from typing import List, NamedTuple, Optional
from urllib import parse
from uuid import uuid4

from django.conf import settings
from django.db import models
from django.http import HttpResponse
from django.utils.functional import cached_property
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from jwcrypto.jwk import JWK
from pylti1p3.deep_link_resource import DeepLinkResource
from pylti1p3.message_launch import MessageLaunch
from pylti1p3.registration import Registration

from .lti_core.constants import ContextRole
from .lti_core.utils import normalize_role, validate_migration_claim


class KeyQuerySet(models.QuerySet):
    """Custom QuerySet for Key objects."""

    def active(self):
        """Excludes inactive keys from the QuerySet."""
        return self.filter(is_active=True)

    def as_jwks(self):
        """Returns active keys as a JWKS."""
        return {"keys": [key.as_jwk() for key in self.active()]}


class BaseKeyManager(models.Manager):
    """Manager for Key objects."""

    def create_from_jwk(self, key):
        """Creates and returns a new Key object using the provided JWK.

        Args:
            key (JWK): The JWK used to create the Key.

        Returns:
            Key: A new Key object.
        """
        return self.create(
            public_key=key.export_to_pem().decode("utf-8"),
            private_key=key.export_to_pem(private_key=True, password=None).decode(
                "utf-8"
            ),
        )

    def generate(self):
        """Creates and returns a new keypair.

        Returns:
            Key: A new Key object.
        """
        return self.create_from_jwk(JWK.generate(kty="RSA", size=2048))


KeyManager = BaseKeyManager.from_queryset(KeyQuerySet)


class Key(models.Model):
    """A keypair for use in LTI integrations.

    Attributes:
        public_key (str): Public key data.
        private_key (str): Private key data.
        is_active (bool): Indicates if the key is present in the JWKS.
        datetime_created (datetime): When the key was created.
        datetime_modified (datetime): When the key was last modified.
    """

    public_key = models.TextField(_("public key"))
    private_key = models.TextField(_("private key"))
    is_active = models.BooleanField(_("is active"), default=True)
    datetime_created = models.DateTimeField(_("created"), default=now, editable=False)
    datetime_modified = models.DateTimeField(_("modified"), auto_now=True)

    objects = KeyManager()

    class Meta:
        verbose_name = _("JWK")
        verbose_name_plural = _("JWKs")
        get_latest_by = "datetime_created"

    def __str__(self):
        return self.as_jwk()["kid"]

    def as_jwk(self):
        """Returns the key as a JWK."""
        jwk_obj = JWK.from_pem(self.public_key.encode("utf-8"))
        public_jwk = json.loads(jwk_obj.export_public())
        public_jwk["alg"] = "RS256"
        public_jwk["use"] = "sig"
        return public_jwk


class LtiRegistrationQuerySet(models.QuerySet):
    """Custom QuerySet for LtiRegistration objects."""

    def active(self):
        """Excludes inactive registrations from the QuerySet."""
        return self.filter(is_active=True)


class LtiRegistration(models.Model):
    """An LTI platform registration.

    Attributes:
        name (str): A display name for this registration
        uuid (uuid4): A unique identifier for registraitons for use in OIDC login init.
        issuer (str): The platform's issuer.
        client_id (str): The client ID assigned to the tool by the platform.
        audience (str): The platforms's OAuth2 audience.
        auth_url (str): The platform's auth login URL.
        token_url (str): The platform's access token retrieval URL.
        keyset_url (str): The platform's JWKS URL.
        is_active (bool): Indicates if the registration is active within the tool.
        public_key (str): Public key data specific to this registration. Only to be used
            when the platform cannot retrieve keys from a tool's JWKS URL.
        private_key (str): Private key data specific to this registration. Only to be
            used when the platform cannot retrieve keys from a tool's JWKS URL.
        datetime_created (datetime): When the registration was created.
        datetime_modified (datetime): When the registration was last modified.
    """

    name = models.CharField(_("name"), max_length=255)
    uuid = models.UUIDField(_("UUID"), default=uuid4)
    issuer = models.CharField(_("issuer"), max_length=255)
    client_id = models.CharField(_("client ID"), max_length=255)
    audience = models.CharField(_("audience"), max_length=255, blank=True)
    auth_url = models.URLField(_("auth URL"))
    token_url = models.URLField(_("access token URL"))
    keyset_url = models.URLField(_("keyset URL"))
    is_active = models.BooleanField(_("is active"), default=True)
    public_key = models.TextField(_("public key"), blank=True)
    private_key = models.TextField(_("private key"), blank=True)
    datetime_created = models.DateTimeField(_("created"), default=now, editable=False)
    datetime_modified = models.DateTimeField(_("modified"), auto_now=True)

    objects = LtiRegistrationQuerySet.as_manager()

    class Meta:
        verbose_name = _("LTI registration")
        verbose_name_plural = _("LTI registrations")
        constraints = [
            models.UniqueConstraint(
                fields=["issuer", "client_id"], name="unique_client_id_per_issuer"
            )
        ]

    def __str__(self):
        return self.name

    @property
    def has_key(self):
        """bool: Indicates if the registration has an assigned keypair."""
        return self.public_key and self.private_key

    def to_registration(self) -> Registration:
        reg = Registration()
        reg.set_auth_login_url(self.auth_url)
        reg.set_auth_token_url(self.token_url)
        reg.set_client_id(self.client_id)
        # reg.set_key_set(key_set)
        reg.set_key_set_url(self.keyset_url)
        reg.set_issuer(self.issuer)
        if self.audience:
            reg.set_auth_audience(self.audience)
        if self.has_key:
            reg.set_tool_public_key(self.public_key)
            reg.set_tool_private_key(self.private_key)
        else:
            key = Key.objects.active().latest()
            reg.set_tool_private_key(key.private_key)
            reg.set_tool_public_key(key.public_key)
        return reg


class LtiPlatformInstance(models.Model):
    """An instance of a LTI learning platform.

    The platform instance claim is described in the LTI spec at
    https://www.imsglobal.org/spec/lti/v1p3/#platform-instance-claim
    """

    issuer = models.CharField(_("issuer"), max_length=255)
    guid = models.CharField(_("GUID"), max_length=255)
    lti1p1_id_on_platform = models.CharField(
        _("LTI 1.1 ID"), max_length=255, blank=True
    )
    contact_email = models.EmailField(_("contact email"), blank=True)
    description = models.TextField(_("description"), blank=True)
    name = models.CharField(_("name"), max_length=500, blank=True)
    url = models.URLField(_("URL"), blank=True)
    product_family_code = models.CharField(
        _("product family code"), max_length=500, blank=True
    )
    version = models.CharField(_("version"), max_length=500, blank=True)

    class Meta:
        verbose_name = _("LTI platform instance")
        verbose_name_plural = _("LTI platform instances")
        constraints = [
            models.UniqueConstraint(
                fields=["issuer", "guid"], name="unique_guid_per_issuer"
            )
        ]

    def __str__(self) -> str:
        return self.name if self.name else self.guid


class LtiDeploymentQuerySet(models.QuerySet):
    """Custom QuerySet for LtiDeployment objects."""

    def active(self):
        """Excludes inactive deployments from the QuerySet."""
        return self.filter(is_active=True)


class LtiDeployment(models.Model):
    """A LTI deployment.

    Deployments are described in the LTI 1.3 core specification at
    https://www.imsglobal.org/spec/lti/v1p3/#tool-deployment.

    Attributes:
        registration (LtiRegistration): The platform registration associated with
            this deployment.
        deployment_id (str): The deployment ID as provided by the platform.
        is_active (bool): Indicates if the deployment has been activated in the tool.
        platform_instance (LtiPlatformInstance): The platform instance associated with
            this deployment.
        datetime_created (datetime): When the deployment was created.
        datetime_modified (datetime): When the deployment was last modified.
    """

    registration = models.ForeignKey(
        LtiRegistration,
        related_name="deployments",
        on_delete=models.CASCADE,
        verbose_name=_("registration"),
    )
    deployment_id = models.CharField(_("deployment ID"), max_length=255)
    is_active = models.BooleanField(_("is active"), default=False)
    platform_instance = models.ForeignKey(
        LtiPlatformInstance,
        related_name="deployments",
        on_delete=models.SET_NULL,
        verbose_name=_("platform instance"),
        null=True,
        blank=True,
    )
    datetime_created = models.DateTimeField(_("created"), default=now, editable=False)
    datetime_modified = models.DateTimeField(_("modified"), auto_now=True)

    objects = LtiDeploymentQuerySet.as_manager()

    class Meta:
        verbose_name = _("LTI deployment")
        verbose_name_plural = _("LTI deployments")
        constraints = [
            models.UniqueConstraint(
                fields=["registration", "deployment_id"],
                name="unique_deployment_id_per_registration",
            )
        ]

    def __str__(self):
        return f"{self.registration}: {self.deployment_id}"


class LtiUser(models.Model):
    """A platform user as described by an LTI launch.

    The user identity claims are described in the LTI spec at
    https://www.imsglobal.org/spec/lti/v1p3/#user-identity-claims
    """

    registration = models.ForeignKey(
        LtiRegistration,
        related_name="users",
        on_delete=models.CASCADE,
        verbose_name=_("registration"),
    )
    sub = models.CharField(_("subject"), max_length=255)
    lti1p1_id_on_platform = models.CharField(
        _("LTI 1.1 ID"), max_length=255, blank=True
    )
    given_name = models.CharField(_("given name"), max_length=500, blank=True)
    family_name = models.CharField(_("family name"), max_length=500, blank=True)
    name = models.CharField(_("name"), max_length=500, blank=True)
    email = models.EmailField(_("email"), blank=True)
    picture_url = models.URLField(_("picture URL"), blank=True)
    auth_user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="lti_users",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name=_("auth user"),
    )
    datetime_created = models.DateTimeField(_("created"), default=now, editable=False)
    datetime_modified = models.DateTimeField(_("modified"), auto_now=True)

    class Meta:
        verbose_name = _("LTI user")
        verbose_name_plural = _("LTI users")
        constraints = [
            models.UniqueConstraint(
                fields=["registration", "sub"], name="unique_sub_per_registration"
            )
        ]

    def __str__(self):
        return self.sub


class LtiContext(models.Model):
    """Describes the context of a LTI launch.

    The context claim are described in the LTI spec at
    https://www.imsglobal.org/spec/lti/v1p3/#context-claim
    """

    deployment = models.ForeignKey(
        LtiDeployment,
        related_name="contexts",
        on_delete=models.CASCADE,
        verbose_name=_("deployment"),
    )
    id_on_platform = models.CharField(_("ID on platform"), max_length=255, blank=True)
    lti1p1_id_on_platform = models.CharField(
        _("LTI 1.1 ID"), max_length=255, blank=True
    )
    label = models.CharField(_("label"), max_length=255, blank=True)
    title = models.CharField(_("title"), max_length=500, blank=True)
    members = models.ManyToManyField(
        LtiUser,
        related_name="contexts",
        through="LtiMembership",
        verbose_name="members",
    )
    is_course_template = models.BooleanField(_("is course template"), default=False)
    is_course_offering = models.BooleanField(_("is course offering"), default=False)
    is_course_section = models.BooleanField(_("is course section"), default=False)
    is_group = models.BooleanField(_("is group"), default=False)
    memberships_url = models.URLField(blank=True)
    lineitems_url = models.URLField(_("lineitems URL"), blank=True)
    can_query_lineitems = models.BooleanField(_("can query lineitems"), default=False)
    can_manage_lineitems = models.BooleanField(_("can manage lineitems"), default=False)
    can_publish_scores = models.BooleanField(_("can publish scores"), default=False)
    can_access_results = models.BooleanField(_("can access results"), default=False)
    datetime_created = models.DateTimeField(_("created"), default=now, editable=False)
    datetime_modified = models.DateTimeField(_("modified"), auto_now=True)

    class Meta:
        verbose_name = _("LTI context")
        verbose_name_plural = _("LTI contexts")
        constraints = [
            models.UniqueConstraint(
                fields=["deployment", "id_on_platform"],
                name="unique_context_id_per_deployment",
            )
        ]

    def __str__(self):
        return self.title if self.title else self.id_on_platform

    def update_memberships(self, member_data: List[dict]):
        """Updates memberships for this context using NRPS data."""
        registration = self.deployment.registration
        for member in member_data:
            user_defaults = {
                "given_name": member.get("given_name"),
                "family_name": member.get("family_name"),
                "name": member.get("name"),
                "email": member.get("email"),
                "picture_url": member.get("picture"),
            }
            user, _created = LtiUser.objects.update_or_create(
                registration=registration,
                sub=member["user_id"],
                defaults={k: v for (k, v) in user_defaults.items() if v is not None},
            )
            member_roles = [normalize_role(role) for role in member["roles"]]
            LtiMembership.objects.update_or_create(
                context=self,
                user=user,
                defaults={
                    "is_administrator": ContextRole.ADMINISTRATOR in member_roles,
                    "is_content_developer": ContextRole.CONTENT_DEVELOPER
                    in member_roles,
                    "is_instructor": ContextRole.INSTRUCTOR in member_roles,
                    "is_learner": ContextRole.LEARNER in member_roles,
                    "is_mentor": ContextRole.MENTOR in member_roles,
                    "is_active": member["status"] == "Active",
                },
            )


class LtiMembership(models.Model):
    """A user's role within a learning context.

    Roles are described in the LTI spec at
    https://www.imsglobal.org/spec/lti/v1p3/#roles-claim
    """

    user = models.ForeignKey(
        LtiUser,
        related_name="memberships",
        on_delete=models.CASCADE,
        verbose_name=_("user"),
    )
    context = models.ForeignKey(
        LtiContext,
        related_name="memberships",
        on_delete=models.CASCADE,
        verbose_name=_("context"),
    )
    is_administrator = models.BooleanField(_("is administrator"), default=False)
    is_content_developer = models.BooleanField(_("is_content_developer"), default=False)
    is_instructor = models.BooleanField(_("is instructor"), default=False)
    is_learner = models.BooleanField(_("is learner"), default=False)
    is_mentor = models.BooleanField(_("is mentor"), default=False)
    is_active = models.BooleanField(_("is active"), default=True)
    datetime_created = models.DateTimeField(_("created"), default=now, editable=False)
    datetime_modified = models.DateTimeField(_("modified"), auto_now=True)

    class Meta:
        verbose_name = _("LTI membership")
        verbose_name_plural = _("LTI memberships")
        constraints = [
            models.UniqueConstraint(
                fields=["user", "context"], name="no_duplicate_memberships"
            )
        ]

    def __str__(self):
        return f"{self.user} in {self.context}"


class LtiResourceLink(models.Model):
    """Represents a resource link in an LTI launch.

    The resource link claim is described in the LTI spec at
    https://www.imsglobal.org/spec/lti/v1p3/#resource-link-claim
    """

    context = models.ForeignKey(
        LtiContext,
        related_name="resource_links",
        on_delete=models.CASCADE,
        verbose_name=_("context"),
    )
    id_on_platform = models.CharField(_("ID on platform"), max_length=255)
    lti1p1_id_on_platform = models.CharField(
        _("LTI 1.1 ID"), max_length=255, blank=True
    )
    title = models.CharField(_("title"), max_length=500, blank=True)
    description = models.TextField(_("description"), blank=True)
    datetime_created = models.DateTimeField(_("created"), default=now, editable=False)
    datetime_modified = models.DateTimeField(_("modified"), auto_now=True)

    class Meta:
        verbose_name = _("LTI resource link")
        verbose_name_plural = _("LTI resource links")
        constraints = [
            models.UniqueConstraint(
                fields=["context", "id_on_platform"],
                name="unique_resource_link_id_per_context",
            )
        ]

    def __str__(self):
        return self.title if self.title else self.id_on_platform


class LtiLineItem(models.Model):
    """Represents an LTI LineItem for assignment scoring.

    The line item scope and methods are described in the LTI Assignment and Grade
    Services 2.0 spec at
    https://www.imsglobal.org/spec/lti-ags/v2p0#line-item-service-scope-and-allowed-http-methods
    """

    context = models.ForeignKey(
        LtiContext,
        related_name="line_items",
        on_delete=models.CASCADE,
        verbose_name=_("context"),
    )
    url = models.URLField(_("URL on platform"), unique=True)
    maximum_score = models.FloatField(_("maximum score"))
    label = models.CharField(_("label"), max_length=255)
    tag = models.CharField(_("tag"), max_length=255, blank=True)
    resource_id = models.CharField(_("resource ID"), max_length=255, blank=True)
    resource_link = models.ForeignKey(
        LtiResourceLink, on_delete=models.SET_NULL, null=True
    )
    start_datetime = models.DateTimeField(_("start date/time"), null=True)
    end_datetime = models.DateTimeField(_("end date/time"), null=True)

    def __str__(self):
        return self.label


class ViewportDimensions(NamedTuple):
    width: int
    height: int


class LtiLaunch:
    """An LTI launch."""

    _lti1p3_message_launch: Optional[MessageLaunch] = None
    _lti1p3_launch_id: Optional[str] = None

    def __init__(self, message_launch: MessageLaunch) -> None:
        launch_id = None
        if message_launch is not None:
            launch_id = message_launch.get_launch_id()
        self._lti1p3_launch_id = launch_id
        self._lti1p3_message_launch = message_launch

    def get_launch_id(self):
        return self._lti1p3_launch_id

    def get_message_launch(self):
        return self._lti1p3_message_launch

    def get_launch_data(self):
        message_launch = self.get_message_launch()
        if message_launch is None:
            return None
        return message_launch.get_launch_data()

    def get_claim(self, claim):
        launch_data = self.get_launch_data()
        if launch_data is None:
            return None
        return launch_data.get(claim)

    @property
    def is_present(self) -> bool:
        return True

    @property
    def is_absent(self) -> bool:
        return False

    @property
    def is_resource_launch(self) -> bool:
        """Indicates if the launch is resource link launch request."""
        message_launch = self.get_message_launch()
        if message_launch is None:
            return False
        return message_launch.is_resource_launch()

    @property
    def is_deep_link_launch(self) -> bool:
        """Indicates if the launch is a deep linking request."""
        message_launch = self.get_message_launch()
        if message_launch is None:
            return False
        return message_launch.is_deep_link_launch()

    @property
    def is_submission_review_launch(self) -> bool:
        """Indicates if the launch is a submission review request."""
        message_launch = self.get_message_launch()
        if message_launch is None:
            return False
        return message_launch.is_submission_review_launch()

    @property
    def is_data_privacy_launch(self) -> bool:
        """Indicates if the launch is a data privacy launch request."""
        message_launch = self.get_message_launch()
        if message_launch is None:
            return False
        return message_launch.is_data_privacy_launch()

    @cached_property
    def registration(self) -> LtiRegistration:
        """The LTI platform registration associated with this launch."""
        tool_conf = self.get_message_launch().get_tool_conf()
        return tool_conf.registration

    @cached_property
    def deployment(self) -> LtiDeployment:
        """The LTI deployment associated with this launch."""
        tool_conf = self.get_message_launch()._tool_config
        if tool_conf.deployment is not None:
            return tool_conf.deployment
        deployment_id = self.get_claim(
            "https://purl.imsglobal.org/spec/lti/claim/deployment_id"
        )
        return LtiDeployment.objects.get(
            registration=self.registration, deployment_id=deployment_id
        )

    @cached_property
    def user(self) -> LtiUser:
        """The LTI user associated with the launch."""
        return LtiUser.objects.get(
            registration=self.registration, sub=self.get_claim("sub")
        )

    @property
    def nrps_claim(self):
        return self.get_claim(
            "https://purl.imsglobal.org/spec/lti-nrps/claim/namesroleservice"
        )

    @property
    def ags_claim(self):
        return self.get_claim("https://purl.imsglobal.org/spec/lti-ags/claim/endpoint")

    @property
    def context_claim(self):
        return self.get_claim("https://purl.imsglobal.org/spec/lti/claim/context")

    @cached_property
    def context(self) -> LtiContext:
        """The LTI context associated with the launch."""
        context_id = self.context_claim["id"] if self.context_claim is not None else ""
        return LtiContext.objects.get(
            deployment=self.deployment, id_on_platform=context_id
        )

    @property
    def roles_claim(self):
        return self.get_claim("https://purl.imsglobal.org/spec/lti/claim/roles")

    @cached_property
    def membership(self) -> LtiMembership:
        """The LTI membership associated with the launch."""
        issuer = self.get_claim("iss")
        aud = self.get_claim("aud")
        client_id = aud[0] if isinstance(aud, list) else aud
        sub = self.get_claim("sub")
        context_id = self.context_claim["id"] if self.context_claim is not None else ""
        return LtiMembership.objects.get(
            user__registration__issuer=issuer,
            user__registration__client_id=client_id,
            user__sub=sub,
            context__id_on_platform=context_id,
            context__deployment=self.deployment,
        )

    @property
    def resource_link_claim(self):
        return self.get_claim("https://purl.imsglobal.org/spec/lti/claim/resource_link")

    @cached_property
    def resource_link(self) -> LtiResourceLink:
        """The LTI resource link associated with the launch."""
        context_id = self.context_claim["id"] if self.context_claim is not None else ""
        return LtiResourceLink.objects.get(
            context__deployment=self.deployment,
            context__id_on_platform=context_id,
            id_on_platform=self.resource_link_claim["id"],
        )

    @property
    def platform_instance_claim(self):
        return self.get_claim("https://purl.imsglobal.org/spec/lti/claim/tool_platform")

    @cached_property
    def platform_instance(self) -> Optional[LtiPlatformInstance]:
        """The LTI platform instance associated with the launch, if present."""
        if self.platform_instance_claim is None:
            return None
        return LtiPlatformInstance.objects.get(
            issuer=self.get_claim("iss"), guid=self.platform_instance_claim["guid"]
        )

    @property
    def launch_presentation_claim(self):
        return self.get_claim(
            "https://purl.imsglobal.org/spec/lti/claim/launch_presentation"
        )

    @property
    def document_target(self) -> Optional[str]:
        """The kind of browser window or frame in which the launch is presented.

        See https://www.imsglobal.org/spec/lti/v1p3/#launch-presentation-claim
        """
        if self.launch_presentation_claim is None:
            return None
        return self.launch_presentation_claim.get("document_target")

    @property
    def dimensions(self) -> Optional[ViewportDimensions]:
        """Width and height of the window or frame in which the launch is presented.

        See https://www.imsglobal.org/spec/lti/v1p3/#launch-presentation-claim
        """
        if self.launch_presentation_claim is None:
            return None
        dimensions = ViewportDimensions(
            self.launch_presentation_claim.get("width"),
            self.launch_presentation_claim.get("height"),
        )
        return dimensions if None not in dimensions else None

    def get_return_url(
        self,
        *,
        lti_errormsg: str = "",
        lti_msg: str = "",
        lti_errorlog: str = "",
        lti_log: str = "",
    ) -> Optional[str]:
        """Constructs a return URL, when supported by the launch."""
        if self.launch_presentation_claim is None:
            return None
        return_url = self.launch_presentation_claim.get("return_url")
        if return_url is None:
            return None
        url_parts = parse.urlsplit(return_url)
        query = dict(
            parse.parse_qsl(url_parts.query)
            + [
                ("lti_errormsg", lti_errormsg),
                ("lti_msg", lti_msg),
                ("lti_errorlog", lti_errorlog),
                ("lti_log", lti_log),
            ]
        )
        return parse.urlunsplit(
            (
                url_parts.scheme,
                url_parts.netloc,
                url_parts.path,
                parse.urlencode([param for param in query.items() if param[1]]),
                url_parts.fragment,
            )
        )

    @property
    def migration_claim(self):
        return self.get_claim("https://purl.imsglobal.org/spec/lti/claim/lti1p1")

    @property
    def lti1p1_consumer_key(self):
        if self.migration_claim is None:
            return None
        return self.migration_claim["oauth_consumer_key"]

    def has_valid_migration_claim(self, lti1p1_secret: str) -> bool:
        """Indicates if the launch has a valid LTI 1.1 migration claim."""
        launch_data = self.get_launch_data()
        if launch_data is None:
            return False
        return validate_migration_claim(launch_data, lti1p1_secret)

    def deep_link_response(self, resources: List[DeepLinkResource]) -> HttpResponse:
        """Creates a deep linking response for this launch."""
        html = self.get_message_launch().get_deep_link().output_response_form(resources)
        return HttpResponse(html)

    def get_custom_claim(self, claim: str) -> Optional[str]:
        """Returns a custom claim value, or None if not present."""
        custom_claims = self.get_claim(
            "https://purl.imsglobal.org/spec/lti/claim/custom"
        )
        return custom_claims.get(claim) if custom_claims is not None else None


class AbsentLtiLaunch:
    """Placeholder for non-LTI launch contexts."""

    @property
    def is_present(self) -> bool:
        return False

    @property
    def is_absent(self) -> bool:
        return True



from typing import Optional

from django.http.request import HttpRequest

from pylti1p3.contrib.django.launch_data_storage.cache import DjangoCacheDataStorage
from pylti1p3.contrib.django.message_launch import DjangoMessageLaunch
from pylti1p3.deployment import Deployment
from pylti1p3.tool_config.abstract import ToolConfAbstract

from .lti_core.constants import AgsScope, ContextRole, ContextType
from .lti_core.utils import normalize_role
from .models import (
    LtiContext,
    LtiDeployment,
    LtiLaunch,
    LtiMembership,
    LtiPlatformInstance,
    LtiRegistration,
    LtiResourceLink,
    LtiUser,
)


def _prepare_deployment(lti_deployment):
    return Deployment().set_deployment_id(lti_deployment.deployment_id)


class DjangoToolConfig(ToolConfAbstract):
    """LTI tool configuration class.

    A registration UUID may be specified on init to address situations
    where client_id isn't included in OIDC initiation params.
    """

    registration_uuid = None
    registration = None
    deployment = None

    def __init__(self, registration_uuid=None):
        super().__init__()
        self.registration_uuid = registration_uuid

    def check_iss_has_one_client(self, iss):
        return False

    def check_iss_has_many_clients(self, iss):
        return True

    def find_registration_by_issuer(self, iss, *args, **kwargs):
        try:
            self.registration = LtiRegistration.objects.active().get(
                uuid=self.registration_uuid, issuer=iss
            )
            return self.registration.to_registration()
        except LtiRegistration.DoesNotExist:
            return None

    def find_registration_by_params(self, iss, client_id, *args, **kwargs):
        lookups = {"issuer": iss, "client_id": client_id}
        if self.registration_uuid is not None:
            lookups.update(uuid=self.registration_uuid)
        try:
            self.registration = LtiRegistration.objects.active().get(**lookups)
            return self.registration.to_registration()
        except LtiRegistration.DoesNotExist:
            return None

    def find_deployment(self, iss, deployment_id):
        try:
            self.deployment = LtiDeployment.objects.get(
                registration__uuid=self.registration_uuid,
                registration__issuer=iss,
                registration__is_active=True,
                deployment_id=deployment_id,
            )
        except LtiDeployment.DoesNotExist:
            self.deployment = LtiDeployment.objects.create(
                registration=self.registration, deployment_id=deployment_id
            )
        return _prepare_deployment(self.deployment)

    def find_deployment_by_params(self, iss, deployment_id, client_id, *args, **kwargs):
        lookups = {
            "registration__issuer": iss,
            "registration__client_id": client_id,
            "registration__is_active": True,
            "deployment_id": deployment_id,
        }
        if self.registration_uuid is not None:
            lookups.update(registration__uuid=self.registration_uuid)
        try:
            self.deployment = LtiDeployment.objects.get(**lookups)
        except LtiDeployment.DoesNotExist:
            self.deployment = LtiDeployment.objects.create(
                registration=self.registration, deployment_id=deployment_id
            )
        return _prepare_deployment(self.deployment)


def get_launch_from_request(
    request: HttpRequest, launch_id: Optional[str] = None
) -> LtiLaunch:
    """Returns the DjangoMessageLaunch associated with a request.

    Optionally, a launch_id may be specified to retrieve the launch from the cache.
    """
    tool_conf = DjangoToolConfig()
    launch_data_storage = DjangoCacheDataStorage()
    if launch_id is not None:
        message_launch = DjangoMessageLaunch.from_cache(
            launch_id, request, tool_conf, launch_data_storage=launch_data_storage
        )
    else:
        message_launch = DjangoMessageLaunch(
            request, tool_conf, launch_data_storage=launch_data_storage
        )
        message_launch.validate()
    return LtiLaunch(message_launch)


def sync_user_from_launch(
    lti_launch: LtiLaunch, lti1p1_secret: Optional[str] = None
) -> LtiUser:
    sub = lti_launch.get_claim("sub")
    user_claims = {
        "given_name": lti_launch.get_claim("given_name"),
        "family_name": lti_launch.get_claim("family_name"),
        "name": lti_launch.get_claim("name"),
        "email": lti_launch.get_claim("email"),
        "picture_url": lti_launch.get_claim("picture"),
    }
    if lti1p1_secret is not None and lti_launch.has_valid_migration_claim(
        lti1p1_secret
    ):
        user_claims["lti1p1_id_on_platform"] = lti_launch.migration_claim.get(
            "user_id", sub
        )
    lti_user, _created = LtiUser.objects.update_or_create(
        registration=lti_launch.registration,
        sub=sub,
        defaults={k: v for k, v in user_claims.items() if v is not None},
    )
    return lti_user


def _get_ags_props(lti_launch: LtiLaunch) -> dict:
    ags_claim = lti_launch.ags_claim
    if ags_claim is None:
        return {}
    ags_scope = ags_claim.get("scope", [])
    return {
        "lineitems_url": ags_claim.get("lineitems", ""),
        "can_query_lineitems": AgsScope.QUERY_LINEITEMS in ags_scope,
        "can_manage_lineitems": AgsScope.MANAGE_LINEITEMS in ags_scope,
        "can_publish_scores": AgsScope.PUBLISH_SCORES in ags_scope,
        "can_access_results": AgsScope.ACCESS_RESULTS in ags_scope,
    }


def sync_context_from_launch(
    lti_launch: LtiLaunch, lti1p1_secret: Optional[str] = None
) -> LtiContext:
    context_claim = {} if lti_launch.context_claim is None else lti_launch.context_claim
    nrps_claim = lti_launch.nrps_claim
    nrps_endpoint = "" if nrps_claim is None else nrps_claim["context_memberships_url"]
    context_types = context_claim.get("type", [])
    defaults = {
        "title": context_claim.get("title", ""),
        "label": context_claim.get("label", ""),
        "is_course_template": (ContextType.COURSE_TEMPLATE in context_types),
        "is_course_offering": (ContextType.COURSE_OFFERING in context_types),
        "is_course_section": (ContextType.COURSE_SECTION in context_types),
        "is_group": (ContextType.GROUP in context_types),
    }
    if nrps_endpoint:
        defaults["memberships_url"] = nrps_endpoint
    defaults.update(_get_ags_props(lti_launch))
    if lti1p1_secret is not None and lti_launch.has_valid_migration_claim(
        lti1p1_secret
    ):
        defaults["lti1p1_id_on_platform"] = lti_launch.migration_claim.get(
            "context_id", context_claim.get("id", "")
        )
    context, _created = LtiContext.objects.update_or_create(
        deployment=lti_launch.deployment,
        id_on_platform=context_claim.get("id", ""),
        defaults=defaults,
    )
    return context


def sync_membership_from_launch(
    lti_launch: LtiLaunch, user: LtiUser, context: LtiContext
) -> LtiMembership:
    roles = [normalize_role(role) for role in lti_launch.roles_claim]
    defaults = {}
    if ContextRole.ADMINISTRATOR in roles:
        defaults["is_administrator"] = True
    if ContextRole.CONTENT_DEVELOPER in roles:
        defaults["is_content_developer"] = True
    if ContextRole.INSTRUCTOR in roles:
        defaults["is_instructor"] = True
    if ContextRole.LEARNER in roles:
        defaults["is_learner"] = True
    if ContextRole.MENTOR in roles:
        defaults["is_mentor"] = True
    membership, _created = LtiMembership.objects.update_or_create(
        user=user, context=context, defaults=defaults
    )
    return membership


def sync_resource_link_from_launch(
    lti_launch: LtiLaunch, context: LtiContext, lti1p1_secret: Optional[str] = None
) -> LtiResourceLink:
    resource_link_claim = {
        k: v for k, v in lti_launch.resource_link_claim.items() if v is not None
    }
    defaults = {
        "title": resource_link_claim.get("title", ""),
        "description": resource_link_claim.get("description", ""),
    }
    if lti1p1_secret is not None and lti_launch.has_valid_migration_claim(
        lti1p1_secret
    ):
        defaults["lti1p1_id_on_platform"] = lti_launch.migration_claim.get(
            "resource_link_id", resource_link_claim["id"]
        )
    resource_link, _created = LtiResourceLink.objects.update_or_create(
        context=context,
        id_on_platform=resource_link_claim["id"],
        defaults=defaults,
    )
    return resource_link


def sync_platform_instance_from_launch(
    lti_launch: LtiLaunch, lti1p1_secret: Optional[str] = None
) -> Optional[LtiPlatformInstance]:
    platform_instance_claim = lti_launch.platform_instance_claim
    if platform_instance_claim is None:
        return None
    defaults = {
        "contact_email": platform_instance_claim.get("contact_email", ""),
        "description": platform_instance_claim.get("description", ""),
        "name": platform_instance_claim.get("name", ""),
        "url": platform_instance_claim.get("url", ""),
        "product_family_code": platform_instance_claim.get("product_family_code", ""),
        "version": platform_instance_claim.get("version", ""),
    }
    if lti1p1_secret is not None and lti_launch.has_valid_migration_claim(
        lti1p1_secret
    ):
        defaults["lti1p1_id_on_platform"] = lti_launch.migration_claim.get(
            "tool_consumer_instance_guid", platform_instance_claim["guid"]
        )
    platform_instance, _created = LtiPlatformInstance.objects.update_or_create(
        issuer=lti_launch.get_claim("iss"),
        guid=platform_instance_claim["guid"],
        defaults=defaults,
    )
    deployment = lti_launch.deployment
    deployment.platform_instance = platform_instance
    deployment.save(update_fields=["platform_instance"])
    return platform_instance


def sync_data_from_launch(
    lti_launch: LtiLaunch, lti1p1_secret: Optional[str] = None
) -> None:
    user = sync_user_from_launch(lti_launch, lti1p1_secret)
    if not lti_launch.is_data_privacy_launch:
        context = sync_context_from_launch(lti_launch, lti1p1_secret)
        sync_membership_from_launch(lti_launch, user, context)
        if not lti_launch.is_deep_link_launch:
            sync_resource_link_from_launch(lti_launch, context, lti1p1_secret)
    sync_platform_instance_from_launch(lti_launch, lti1p1_secret)



# Generated by Django 4.0.2 on 2022-10-20 16:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('lti_tool', '0003_lticontext_memberships_url_ltimembership_is_active'),
    ]

    operations = [
        migrations.AddField(
            model_name='lticontext',
            name='can_access_results',
            field=models.BooleanField(default=False, verbose_name='can access results'),
        ),
        migrations.AddField(
            model_name='lticontext',
            name='can_manage_lineitems',
            field=models.BooleanField(default=False, verbose_name='can manage lineitems'),
        ),
        migrations.AddField(
            model_name='lticontext',
            name='can_publish_scores',
            field=models.BooleanField(default=False, verbose_name='can publish scores'),
        ),
        migrations.AddField(
            model_name='lticontext',
            name='can_query_lineitems',
            field=models.BooleanField(default=False, verbose_name='can query lineitems'),
        ),
        migrations.AddField(
            model_name='lticontext',
            name='lineitems_url',
            field=models.URLField(blank=True, verbose_name='lineitems URL'),
        ),
    ]



import pytest

from lti_tool import factories, models


@pytest.mark.django_db
class TestLtiRegistrationQuerySet:
    """Tests for LtiRegistrationQuerySet."""

    def test_active(self):
        factories.LtiRegistrationFactory.create_batch(2)
        inactive_registration = factories.LtiRegistrationFactory(is_active=False)
        active_registrations = models.LtiRegistration.objects.active()
        assert inactive_registration not in active_registrations
        assert len(active_registrations) == 2


@pytest.mark.django_db
class TestLtiRegistration:
    """Tests for the LtiRegistration model."""

    def test_str(self):
        registration = factories.LtiRegistrationFactory(name="LTI Registration")
        assert str(registration) == "LTI Registration"


@pytest.mark.django_db
class TestLtiDeploymentQuerySet:
    """Tests for LtiDeploymentQuerySet."""

    def test_active(self):
        factories.LtiDeploymentFactory.create_batch(2, is_active=True)
        inactive_deployment = factories.LtiDeploymentFactory()
        active_deployments = models.LtiDeployment.objects.active()
        assert inactive_deployment not in active_deployments
        assert len(active_deployments) == 2


@pytest.mark.django_db
class TestLtiDeployment:
    """Tests for the LtiDeployment model."""

    def test_str(self):
        deployment = factories.LtiDeploymentFactory(
            registration__name="Registration", deployment_id="a-deployment-id"
        )
        assert str(deployment) == "Registration: a-deployment-id"


@pytest.mark.django_db
class TestLtiPlatformInstance:
    """Tests for the LtiPlatformInstance model."""

    @pytest.mark.parametrize(
        ("guid", "name", "result"), [("guid", "Name", "Name"), ("guid", "", "guid")]
    )
    def test_str(self, guid, name, result):
        platform_instance = factories.LtiPlatformInstanceFactory(name=name, guid=guid)
        assert str(platform_instance) == result


@pytest.mark.django_db
class TestLtiUser:
    """Tests for the LtiUser model."""

    def test_str(self):
        user = factories.LtiUserFactory(sub="abc123")
        assert str(user) == "abc123"


@pytest.mark.django_db
class TestLtiContext:
    """Tests for the LtiContext model."""

    @pytest.mark.parametrize(
        ("id", "title", "result"),
        [("ctx-4444", "", "ctx-4444"), ("ctx-4444", "Title", "Title")],
    )
    def test_str(self, id, title, result):
        context = factories.LtiContextFactory(id_on_platform=id, title=title)
        assert str(context) == result

    def test_update_memberships(self):
        context = factories.LtiContextFactory()
        member_data = [
            {
                "status": "Active",
                "name": "Jane Q. Public",
                "picture": "https://platform.example.edu/jane.jpg",
                "given_name": "Jane",
                "family_name": "Doe",
                "email": "jane@platform.example.edu",
                "user_id": "user_1",
                "roles": ["Instructor"],
            },
            {
                "status": "Inactive",
                "user_id": "user_2",
                "roles": ["http://purl.imsglobal.org/vocab/lis/v2/membership#Learner"],
            },
        ]
        context.update_memberships(member_data)

        member_1 = models.LtiMembership.objects.get(context=context, user__sub="user_1")
        member_2 = models.LtiMembership.objects.get(context=context, user__sub="user_2")

        assert context.members.count() == 2

        assert not member_1.is_administrator
        assert not member_1.is_content_developer
        assert member_1.is_instructor
        assert not member_1.is_learner
        assert not member_1.is_mentor
        assert member_1.is_active
        assert member_1.user.name == "Jane Q. Public"
        assert member_1.user.given_name == "Jane"
        assert member_1.user.family_name == "Doe"
        assert member_1.user.picture_url == "https://platform.example.edu/jane.jpg"
        assert member_1.user.email == "jane@platform.example.edu"

        assert not member_2.is_administrator
        assert not member_2.is_content_developer
        assert not member_2.is_instructor
        assert member_2.is_learner
        assert not member_2.is_mentor
        assert not member_2.is_active


@pytest.mark.django_db
class TestLtiMembership:
    """Tests for the LtiMembership model."""

    def test_str(self):
        membership = factories.LtiMembershipFactory(
            user__sub="user1234", context__title="Math"
        )
        assert str(membership) == "user1234 in Math"


@pytest.mark.django_db
class TestLtiResourceLink:
    """Tests for the LtiResourceLink model."""

    @pytest.mark.parametrize(
        ("id", "title", "result"),
        [("abc123", "Title", "Title"), ("abc123", "", "abc123")],
    )
    def test_str(self, id, title, result):
        resource_link = factories.LtiResourceLinkFactory(id_on_platform=id, title=title)
        assert str(resource_link) == result


@pytest.mark.django_db
class TestLtiLaunch:
    """Tests for the LtiLaunch object."""

    def test_membership_with_duplicate_context_ids(self, monkeypatch):
        previous_membership = factories.LtiMembershipFactory()
        registration = previous_membership.user.registration
        new_membership = factories.LtiMembershipFactory(
            context__deployment__registration=registration,
            context__id_on_platform=previous_membership.context.id_on_platform,
            user=previous_membership.user,
        )
        new_deployment = new_membership.context.deployment
        launch_data = {
            "iss": registration.issuer,
            "aud": registration.client_id,
            "sub": new_membership.user.sub,
            "https://purl.imsglobal.org/spec/lti/claim/deployment_id": (
                new_deployment.deployment_id
            ),
            "https://purl.imsglobal.org/spec/lti/claim/context": {
                "id": new_membership.context.id_on_platform,
            },
        }
        monkeypatch.setattr(
            models.LtiLaunch, "get_launch_data", lambda self: launch_data
        )
        monkeypatch.setattr(models.LtiLaunch, "deployment", new_deployment)
        lti_launch = models.LtiLaunch(None)
        assert lti_launch.membership == new_membership

    def test_membership_with_aud_array(self, monkeypatch):
        membership = factories.LtiMembershipFactory()
        deployment = membership.context.deployment
        registration = deployment.registration
        launch_data = {
            "iss": registration.issuer,
            "aud": [registration.client_id],
            "sub": membership.user.sub,
            "https://purl.imsglobal.org/spec/lti/claim/deployment_id": (
                deployment.deployment_id
            ),
            "https://purl.imsglobal.org/spec/lti/claim/context": {
                "id": membership.context.id_on_platform,
            },
        }
        monkeypatch.setattr(
            models.LtiLaunch, "get_launch_data", lambda self: launch_data
        )
        monkeypatch.setattr(models.LtiLaunch, "deployment", deployment)
        lti_launch = models.LtiLaunch(None)
        assert lti_launch.membership == membership
